#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: Cache.pm,v 1.54 2004/09/07 13:14:19 fuktommy Exp $
#

package Shingetsu::Cache;
use Shingetsu::AttachCache;
use Shingetsu::CacheStat;
use Shingetsu::Config;
use Shingetsu::Node;
use Shingetsu::NodeList;
use Shingetsu::Signature;
use Shingetsu::Util;

#
# configration
#
my $datadir  = $Shingetsu::Config::datadir;
my $cachedir = $Shingetsu::Config::cachedir;

#
# add new data cache
#
sub addData($@) {
	my($file, @newdata) = @_;
	my @data = ();
	if (open IN, "$datadir/$file.dat") {
		@data = <IN>;
		chomp @data;
		close IN;
	}
	Shingetsu::Util::lock();
	open OUT, "> $datadir/$file.dat.tmp" or die "$datadir/$file.tmp: open failed. $!";
	my $str = "";
	@data = sort (@data, @newdata);
	my $records = 0;
	local $_;
	foreach (@data) {
		if ($_ ne $str) {
			print OUT "$_\n";
			$str = $_;
			$records++;
		}
	}
	close OUT;
	rename "$datadir/$file.dat.tmp", "$datadir/$file.dat" or
			die "$datadir/$file: rename failed. $!";
	Shingetsu::Util::unlock();

	foreach (@newdata) {
		my %rec = Shingetsu::Util::rec($_);
		if (($rec{remove_id}) && ($rec{remove_stamp})
			&& (grep $_ eq "remove_stamp", split(/,/, $rec{target}))
			&& (grep $_ eq "remove_id", split(/,/, $rec{target}))
			&& Shingetsu::Signature::include(Shingetsu::Signature::pubkey2trip($rec{pubkey}))
			&& Shingetsu::Signature::check(%rec)) {
			Shingetsu::Cache::removeRecord($file, $rec{remove_stamp}, $rec{remove_id});
		}
		Shingetsu::AttachCache::addRecord($file, $_, %rec);
	}

	local @_ = split /<>/, $data[-1];
	$laststamp = $_[0];
	my %stat = Shingetsu::CacheStat::list();
	$stat{$file} = {
			stamp	=> $laststamp,
			records	=> $records,
			size	=> (-s "$datadir/$file.dat"),
		};
	Shingetsu::CacheStat::sync(%stat);
}

#
# get appointed data
#
sub getData($$$$) {
	my($file, $stamp, $id, $node) = @_;
	my @newdata = $node->talk("/get/$file/$stamp/$id");
	@newdata = grep /^$stamp<>[0-9A-Za-z]+<>/, @newdata;
	@newdata = Shingetsu::Util::md5check(@newdata);
	addData($file, @newdata);
	return (@newdata != 0);
}

#
# get data in region
#
sub getRegion($$) {
	my($file, $node) = @_;
	local $_;
	my %stat = Shingetsu::CacheStat::list();
	my $begin = 0;
	if ($stat{$file}{stamp} && (-s "$datadir/$file.dat")) {
		$begin = $stat{$file}{stamp};
	}
	my $now = time - $Shingetsu::Config::syncsafety * 60*60;
	$begin = $now if ($now < $begin);
	my @newdata = ();
	if ($begin <= 0) {
		@newdata = $node->talk("/get/$file/0-");
		@newdata = Shingetsu::Util::md5check(@newdata);
	} else {
		my @head = $node->talk("/head/$file/$begin-");
		chomp @head;
		my @tmp = ();
		my $h;
		my @data = ();
		if (open IN, "$cachedir/$file/head.cache") {
			@data = <IN>;
			close IN;
			chomp @data;
		}
		foreach $h (@head) {
			next if (grep /^$h(<>|$)/, @data);
			my($stamp, $id) = split /<>/, $h;
			my @buf = $node->talk("/get/$file/$stamp/$id");
			@buf = grep /^$stamp<>[0-9A-Za-z]+<>/, @buf;
			@buf = Shingetsu::Util::md5check(@buf);
			Shingetsu::Util::lock();
			open OUT, ">> $datadir/$file.dat" or die "$datadir/$file: open failed. $!";
			print OUT join("\n", @buf), "\n";
			close OUT;
			Shingetsu::Util::unlock();
			foreach (@buf) {
				Shingetsu::AttachCache::addRecord($file, $_);
			}
		}
	}
	addData($file, @newdata);
}

#
# search node which has the file
# args: file
#
sub search($) {
	my($file) = @_;
	local $_;
	my @node = Shingetsu::NodeList::all();
	my $myself = Shingetsu::NodeList::myself();
	$myself = ($myself)? $myself->toString(): "";
	my %done = ();
	my $found = 0;
	my $count = 0;
	while (@node) {
		$node = shift @node;
		my @buf = $node->talk("/have/$file");
		if ((defined $buf[0]) && ($buf[0] eq "YES")) {
			$found = 1;
			last;
		} elsif ($count++ > $Shingetsu::Config::searchDepth) {
			last;
		} else {
			for (my $i=0; $i<$Shingetsu::Config::askNode; $i++) {
				@buf = $node->talk("/node");
				if ((! defined $buf[0]) || ($buf[0] eq $myself)) {
					next;
				}
				unless ($done{$node->toString()}++ > $Shingetsu::Config::retrySearch) {
					push @node, (new Shingetsu::Node($buf[0]));
				}
			}
		}
	}
	if ($found) {
		return $node;
	} else {
		return undef;
	}
}

#
# sync all files
#
sub sync {
	my @file = glob "$datadir/*.dat";
	local $_;
	for ($_=@file; $_>0; $_--) {
		my $t = int rand $_;
		($file[$_-1], $file[$t]) = ($file[$t], $file[$_-1]);
	}
	foreach (@file) {
		next if ((-s $_) > $Shingetsu::Config::filelimit*1024*1024);
		next if (! (-e $_));
		s|$datadir/||;	s|\.dat$||;
		my $node = search($_) and getRegion($_, $node);
	}
	warn "Cache::sync() finished\n";
}

#
# tell update other nodes
# if $node==null then $node := myself
#
sub tellupdate {
	my ($file, $stamp, $id, $node) = @_;
	my @node = Shingetsu::NodeList::all();
	local $_;
	my $pid;
	if ($pid = fork) {
		return;
	} else {
		local $SIG{TERM} = $SIG{HUP} = "DEFAULT";
		open STDOUT, "> /dev/null";
		my $xnode;
		if (defined $node) {
			$xnode = $node->toXString();
		} else {
			$xnode = ":$ENV{SERVER_PORT}$Shingetsu::Config::path";
			$xnode =~ s|/|+|g;
		}
		foreach (@node) {
			$_->talk("/update/$file/$stamp/$id/$xnode");
		}
		exit;
	}
}

#
# make new files writable
#
sub newFiles() {
	local $_;
	foreach (glob("$datadir/*.new")) {
		my $new = $_;
		s/\.new$//;
		rename $new, $_;
		my $file = $_;
		$file =~ s|^$datadir/||;
		$file =~ s|\.dat$||;
		my %stat = Shingetsu::CacheStat::list();
		$stat{$file} = {
                        stamp   => 0,
                        records => 1,
                        size	=> (-s "$datadir/$file.dat"),
                };
		Shingetsu::CacheStat::sync(%stat);
	}
}

#
# remove record
#
sub removeRecord($$$) {
	my($file, $stamp, $id) = @_;
	open IN, "$datadir/$file.dat" or return undef;
	Shingetsu::Util::lock();
	open OUT, "> $datadir/$file.dat.tmp" or die "$datadir/$file.dat.tmp: failed to open. $!";
		while (<IN>) {
		if (/^$stamp<>$id<>/) {
			print OUT "$stamp<>$id\n";
		} else {
			print OUT;
		}
	}
	close OUT;
	close IN;
	rename "$datadir/$file.dat.tmp", "$datadir/$file.dat" or die "$datadir/$file: rename failed. $!";
	Shingetsu::Util::unlock();

	Shingetsu::AttachCache::removeRecord($file, $stamp, $id) or return undef;
	return 1;
}

#
# remove note history
#
sub cleanHistory() {
	my $savehistory = $Shingetsu::Config::savehistory;
	return if ($savehistory < 0);
	local $_;
	my $file;
	my $now = time;
	foreach $file (glob "$datadir/note_*.dat") {
		$file =~ s|$datadir/||;
		$file =~ s|\.dat$||;
		open IN, "$cachedir/$file/head.cache" or next;
		my @head = <IN>;
		close IN;
		foreach (@head[0..$#head-1]) {
			chomp;
			next unless (/<>$/);
			@_ = split /<>/;
			next if (@_ < 2);
			if ($_[0] + $savehistory*60*60 < $now) {
				removeRecord($file, $_[0], $_[1])
			}
		}
	}
}

#
# remove header
#
sub removeHeader($$$) {
	my($file, $stamp, $id) = @_;
	local $_;
	open IN, $file or return undef;
	Shingetsu::Util::lock();
	open OUT, "> $file.tmp" or die "$file.tmp: failed to open. $!";
	while (<IN>) {
		unless (/^$stamp<>$id/) {
			print OUT;
		}
	}
	close OUT;
	close IN;
	rename "$file.tmp", $file or die "$file: rename failed. $!";
	Shingetsu::Util::unlock();
	return 1;
}

#
# clean old header
#
sub cleanHeader() {
	my $saveheader = $Shingetsu::Config::saveheader;
	return if ($saveheader < 0);
	local $_;
	my $file;
	my $now = time;
	foreach $file (glob "$datadir/*.dat") {
		$file =~ s|$datadir/||;
		$file =~ s|\.dat$||;
		open IN, "$cachedir/$file/head.cache" or next;
		my @buf = <IN>;
		close IN;
		foreach (@buf) {
			@_ = split /<>/;
			if ((@_ >= 2) && (! /<>$/) && ($_[0] + $saveheader*60*60 < $now)) {
				my $f;
				foreach $f ("$datadir/$file.dat",
					    "$cachedir/$file/body.cache", "$cachedir/$file/head.cache") {
					removeHeader($f, $_[0], $_[1]);
				}
			}

		}
	}
	warn "Cache::cleanHeader() finished\n";
}

1;
